
// a simple driver for the multirnafold

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iostream>
//#include <time.h>

// include the multirnafold library
#include "simfold.h"
#include "pairfold.h"
#include "multifold.h"
#include "globals.h"
//#include "timer.h"

using namespace std;

int main (int argc, char *argv[])
{
    char sequences[MAXNUMSEQ][MAXSLEN];
    char structure[MAXSUBSTR*MAXSLEN];
    double energy;
    int i;
    int num_sequences;
        
    if (argc < 2)
    {
        printf ("Usage: %s <sequence1> [<sequence2> [sequence3] ...]\n", argv[0]);
        return 0;
    }    
        
    num_sequences = argc-1;
    for (i = 1; i < argc; i++)
    {
        strcpy (sequences[i-1], argv[i]);
    }
    num_sequences = argc-1;    
    
    // Before calling any function in the library, you have to initialize config_file, dna_or_rna, temperature
    //     and to call the function init_data, which loads the thermodynamic parameters into memory
    
    // configuration file
    char config_file[200] = "params/multirnafold.conf";

    // what to fold: RNA or DNA
    int dna_or_rna = RNA;

    // temperature: any integer or real number between 0 and 100
    // represents degrees Celsius
    double temperature = 37; 
    
    // initialize the thermodynamic parameters
    // call init_data only once for the same dna_or_rna and same temperature
    // if one of them changes, call init_data again   
    init_data (argv[0], config_file, dna_or_rna, temperature);        
        
    if (num_sequences == 1)
    {
        energy = simfold (sequences[0], structure);
        printf ("%s\n%s  %.2lf\n", sequences[0], structure, energy);
    }
    else if (num_sequences == 2)
    {
        energy = pairfold_mfe (sequences[0], sequences[1], structure);
        printf ("%s %s\n%s  %.2lf\n", sequences[0], sequences[1], structure, energy);        
    }
    else
    {
        energy = multifold_ordered (num_sequences, sequences, structure);
        for (i=0; i < num_sequences; i++)
        {
            printf ("%s ", sequences[i]);
        }
        printf ("\n");
        printf ("%s  %.2lf\n", structure, energy);        
    }
    return 0;
}
    


