/***************************************************************************
                          p_multi_loop.cpp  -  description
                             -------------------
    begin                : Mon Apr 15 2002
    copyright            : (C) 2002 by Mirela Andronescu
    email                : andrones@cs.ubc.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <malloc.h>

#include "externs.h"
#include "common.h"
#include "p_multi_loop.h"
#include "pairfold.h"
#include "p_energy_matrix.h"

p_multi_loop::p_multi_loop (int *seq, int link, int length)
// The constructor
{
    int i;
    sequence = seq;
    seqlen = length;
    this->V = NULL;
    this->link = link;
    
    index = (int*) malloc (sizeof(int)*length);
    int total_length;
    total_length = (length *(length+1))/2;
    index[0] = 0;
    for (i=1; i < length; i++)
        index[i] = index[i-1]+length-i+1;

    WM = new int [total_length];
    if (WM == NULL) giveup ("Cannot allocate memory", "p_multi_loop");
    for (i=0; i < total_length; i++) 
      {
        WM[i] = INF;
      }
                                    
    WM_link = new int [total_length];
    if (WM_link == NULL) giveup ("Cannot allocate memory", "p_multi_loop");
    for (i=0; i < total_length; i++) 
      {
        WM_link[i] = INF;
      }        
}



p_multi_loop::~p_multi_loop ()
// The destructor
{
    delete [] index;
    delete [] WM;
    delete [] WM_link;
}


void p_multi_loop::compute_energy_WM (int j)
// computes the free energy of a fragment of a regular multi-loop closed by (i,j)
{
    int i;
    int tmp;
    int ij, iplus1j, ijminus1, ik, kplus1j;

    for (i=j-TURN-1; i>=0; i--)
    {
        ij = index[i]+j-i;
        iplus1j = index[i+1]+j-i-1;
        ijminus1 = index[i]+j-1-i;


        tmp = V->get_energy(i,j) +
                   AU_penalty (sequence[i], sequence[j]) +
                   misc.multi_helix_penalty;
        if (tmp < WM[ij])
          {
            WM[ij] = tmp;
          }

        if ( i != link)
        {
            tmp = V->get_energy(i+1,j) +
                  AU_penalty (sequence[i+1], sequence[j]) +
                  dangle_bot [sequence[j]]
                             [sequence[i+1]]
                             [sequence[i]] +
                  misc.multi_helix_penalty +
                  misc.multi_free_base_penalty;
            if (tmp < WM[ij])
            {
                WM[ij] = tmp;
            }
        }

        if (j != link+1)

        {
            tmp = V->get_energy(i,j-1) +
                  AU_penalty (sequence[i], sequence[j-1]) +
                  dangle_top [sequence [j-1]]
                             [sequence [i]]
                             [sequence [j]] +
                  misc.multi_helix_penalty +
                  misc.multi_free_base_penalty;
            if (tmp < WM[ij])
            {
                WM[ij] = tmp;
            }
        }

        if (i != link && j != link+1)
        {
            tmp = V->get_energy(i+1,j-1) +
                  AU_penalty (sequence[i+1], sequence[j-1]) +
                  dangle_bot [sequence[j-1]]
                             [sequence[i+1]]
                             [sequence[i]] +
                  dangle_top [sequence [j-1]]
                             [sequence [i+1]]
                             [sequence [j]] +
                  misc.multi_helix_penalty +
                  2*misc.multi_free_base_penalty;
            if (tmp < WM[ij])
            {
                WM[ij] = tmp;
            }
        }
        if (link != i)
          {
            tmp = WM[iplus1j] + misc.multi_free_base_penalty;            
            if (tmp < WM[ij])
              {
                WM[ij] = tmp;
              }
          }
        if (link != j-1)
          {
            tmp = WM[ijminus1] + misc.multi_free_base_penalty;
            if (tmp < WM[ij])
              {
                WM[ij] = tmp;
              }
          }

        for (int k=i; k < j; k++)
          {
            if (link == k) continue;
            ik = index[i]+k-i;
            kplus1j = index[k+1]+j-k-1;
            tmp = WM[ik] + WM[kplus1j];
            if (tmp < WM[ij])
              {
                WM[ij] = tmp;
              }
          }

    }
}


void p_multi_loop::compute_energy_WM_link (int j)
// computes the free energy of a fragment of a special multi-loop closed by (i,j)
{
    int i;
    int tmp;
    int ij, iplus1j, ijminus1, ik, kplus1j;

    for (i=j-TURN-1; i>=0; i--)
    {
            
        ij = index[i]+j-i;
        iplus1j = index[i+1]+j-i-1;
        ijminus1 = index[i]+j-1-i;

        // added Jan 26, 2004
        // if i <= b < j, return infinity
        
        if (i <= link && link < j)
        {
            WM_link[ij] = INF;
            continue;
        }                                
        
        // calculate WM[i][j]
        tmp = V->get_energy(i,j) + AU_penalty (sequence[i], sequence[j]);
        if (tmp < WM_link[ij])
          {
            WM_link[ij] = tmp;
          }

        tmp = V->get_energy(i+1,j) + AU_penalty (sequence[i+1], sequence[j]) + 
          dangle_bot [sequence[j]][sequence[i+1]][sequence[i]];
        if (tmp < WM_link[ij])
          {
            WM_link[ij] = tmp;
          }

        tmp = V->get_energy(i,j-1) + AU_penalty (sequence[i], sequence[j-1]) + 
          dangle_top [sequence [j-1]][sequence [i]][sequence [j]];
        if (tmp < WM_link[ij])
          {
            WM_link[ij] = tmp;
          }

        tmp = V->get_energy(i+1,j-1) +
              AU_penalty (sequence[i+1], sequence[j-1]) + 
                   dangle_bot [sequence[j-1]]
                              [sequence[i+1]]
                              [sequence[i]] +
                   dangle_top [sequence [j-1]]
                              [sequence [i+1]]
                              [sequence [j]];
        if (tmp < WM_link[ij])
          {
            WM_link[ij] = tmp;
          }

        tmp = WM_link[iplus1j];
        if (tmp < WM_link[ij])
          {
            WM_link[ij] = tmp;
          }

        tmp = WM_link[ijminus1];
        if (tmp < WM_link[ij])
          {
            WM_link[ij] = tmp;
          }

        for (int k=i; k < j; k++)
          {
            ik = index[i]+k-i;
            kplus1j = index[k+1]+j-k-1;
            tmp = WM_link[ik] + WM_link[kplus1j];
            if (tmp < WM_link[ij])
              {
                WM_link[ij] = tmp;
              }
          }

    }
}


int p_multi_loop::compute_energy_regular (int i, int j)
// computes the free energy of a regular multi-loop closed by (i,j)
{
    int min = INF, tmp, k;
    int iplus1k;
    int kplus1jminus1;
    int iplus2k;
    int kplus1jminus2;

    if (i == link || j-1 == link) 
      return INF;

    for (k = i+TURN+1; k <= j-TURN-2; k++)
      {
        iplus1k = index[i+1] + k -i-1;
        kplus1jminus1 = index[k+1] + j-1 -k-1;
        iplus2k = index[i+2] + k -i-2;
        kplus1jminus2 = index[k+1] + j-2 -k-1;
        
        if (link != k)
          {
            tmp = WM[iplus1k] + WM[kplus1jminus1];
            if (tmp < min)
              min = tmp;
          }
        if (link != i+1 && link != k)
          {
            tmp = WM[iplus2k] + WM[kplus1jminus1] + 
              dangle_top [sequence [i]]
              [sequence [j]]
              [sequence [i+1]] + 
              misc.multi_free_base_penalty;
            if (tmp < min)
              min = tmp;
          }
        if (link != k && link != j-2)
          {
            tmp = WM[iplus1k] + WM[kplus1jminus2] + 
              dangle_bot [sequence[i]]
              [sequence[j]]
              [sequence[j-1]] + 
              misc.multi_free_base_penalty;
            if (tmp < min)
              min = tmp;
          }
        if (link != i+1 && link != k && link != j-2)
          {
            tmp = WM[iplus2k] + WM[kplus1jminus2] + 
              dangle_top [sequence [i]]
              [sequence [j]]
              [sequence [i+1]] +  
              dangle_bot [sequence[i]]
              [sequence[j]]
              [sequence[j-1]] + 
              2 * misc.multi_free_base_penalty;
              
            // moved this inside the brackets on Jan 21, 2005  
            if (tmp < min)
              min = tmp;              
          }

      }
    
                          
    min += misc.multi_helix_penalty + misc.multi_offset +
           AU_penalty (sequence[i], sequence[j]);
    return min;
}



int p_multi_loop::compute_energy_link (int i, int j)
// computes the free energy of a special multi-loop closed by (i,j)
{
    int k;
    int min = INF, tmp;
    int iplus1b;
    int iplus2b;
    int bplus1jminus1;
    int bplus1jminus2;
    int bplus1k, kplus1jminus1, kplus1jminus2, kplus1b, iplus1k, iplus2k ;


    iplus1b = index[i+1]+link-i-1;
    iplus2b = index[i+2]+link-i-2;
    bplus1jminus1 = index[link+1]+j-1-link-1;
    bplus1jminus2 = index[link+1]+j-2-link-1;        
    
    if (link < i || link >= j)
        return INF;

    if (link > i+2 && link +1 < j-2)
      {
        tmp = WM_link[iplus1b] + WM_link[bplus1jminus1];      
        if (tmp < min)
          {
            min = tmp;
          }
        
        tmp = WM_link[iplus2b] + WM_link[bplus1jminus1];
        tmp += dangle_top [sequence [i]]
          [sequence [j]]
          [sequence [i+1]];
        if (tmp < min)
          {
            min = tmp;
          }
        
        tmp = WM_link[iplus1b] + WM_link[bplus1jminus2] +
          dangle_bot [sequence[i]]
          [sequence[j]]
          [sequence[j-1]];
        if (tmp < min)
          {
            min = tmp;
          }
        
        tmp = WM_link[iplus2b] + WM_link[bplus1jminus2]+
          dangle_top [sequence[i]][sequence[j]][sequence[i+1]] +
          dangle_bot [sequence[i]][sequence[j]][sequence[j-1]];
        if (tmp < min)
          {
            min = tmp;
          }
      }

    for (k = link+TURN+2; k <= j-TURN-2; k++)
      {
        bplus1k = index[link+1] + k - link-1;
        kplus1jminus1 = index[k+1] + j-1 - k-1;
        tmp = WM_link[bplus1k] + WM_link[kplus1jminus1];
        if (link != i) 
          tmp += dangle_top[sequence[i]][sequence[j]][sequence[i+1]];
        if (tmp < min)
          {
            min = tmp;
          }
      }
    
    for (k = link+TURN+2; k <= j-TURN-3; k++)
      {
        bplus1k = index[link+1] + k - link-1;
        kplus1jminus2 = index[k+1] + j-2 - k-1;
        tmp = WM_link[bplus1k] + WM_link[kplus1jminus2] +
          dangle_bot [sequence[i]][sequence[j]][sequence[j-1]];
        if (link != i) 
          tmp += dangle_top[sequence[i]][sequence[j]][sequence[i+1]];
        if (tmp < min)
          {
            min = tmp;
          }
      }

    for (k = i+TURN+2; k <= link-TURN-1; k++)
      {
        iplus1k = index[i+1] + k - i-1;
        kplus1b = index[k+1] + link - k-1;
        tmp = WM_link[iplus1k] + WM_link[kplus1b];
        if (link != j-1) 
          tmp += dangle_bot [sequence[i]][sequence[j]][sequence[j-1]];
        if (tmp < min)
          {
            min = tmp;
          }
      }

    for (k = i+TURN+3; k <= link-TURN-1; k++)
      {
        iplus2k = index[i+2] + k - i-2;
        kplus1b = index[k+1] + link - k-1;
        tmp = WM_link[iplus2k] + WM_link[kplus1b] + 
          dangle_top[sequence[i]][sequence[j]][sequence[i+1]];
        if (link != j-1) 
          tmp += dangle_bot [sequence[i]][sequence[j]][sequence[j-1]];
        if (tmp < min)
          {
            min = tmp;
          }
      }


    min += misc.intermolecular_initiation + AU_penalty (sequence[i], sequence[j]);
    return min;        
}

